# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['plkit']

package_data = \
{'': ['*']}

install_requires = \
['cmdy',
 'diot',
 'optuna>=2.0.0,<3.0.0',
 'pytorch-lightning',
 'rich>=8.0.0,<9.0.0',
 'torch>=1.0.0,<2.0.0',
 'tqdm==4.48.0']

setup_kwargs = {
    'name': 'plkit',
    'version': '0.0.7',
    'description': 'A wrapper for pytorch-lightning to make you write even less code',
    'long_description': '# plkit\n\nA wrapper of [pytorch-lightning][1] that makes you write even less code.\n\n[![pypi version][3]][4] [![github tag][5]][6] [![codacy quality][7]][8] [![coverage][8]][9] ![docs][10] ![building][11]\n\n## Installation\n\n```\npip install -U plkit\n```\n\n## Principles\n\n- Being compatible with pytorch-lightning\n- Using configurations instead of coding if possible\n\n## Features\n- Compatible with `pytorch-lightning`\n- Exposed terminal logger\n- Better formatted warnings and errors from `pytorch-lightning`\n- Trainer from a dictionary (not only from an `ArgumentParser` or a `Namespace`)\n- Data module with automatic split for train, val and test datasets\n- Auto loss function and optimizer\n- Builtin measurements (using `pytorch-lightning.metrics`)\n- Optuna integration\n\n## Usage\n\n### From pytorch-lightning\'s minimal example\n\n```python\n"""A minimal example for plkit"""\n\nfrom pathlib import Path\nimport torch\nfrom torchvision import transforms\nfrom torchvision.datasets import MNIST\nfrom plkit import Module, DataModule, run\n\nclass Data(DataModule):\n\n    def data_reader(self):\n        return MNIST(Path(__file__).parent / \'data\', train=True,\n                     download=True, transform=transforms.ToTensor())\n\nclass LitClassifier(Module):\n\n    def __init__(self, config):\n        super().__init__(config)\n        self.l1 = torch.nn.Linear(28 * 28, 10)\n\n    def forward(self, x):\n        return torch.relu(self.l1(x.view(x.size(0), -1).float()))\n\n    def training_step(self, batch, _):\n        x, y = batch\n        loss = self.loss_function(self(x), y)\n        return {\'loss\': loss}\n\nif __name__ == \'__main__\':\n    configuration = {\n        \'gpus\': 1,\n        \'data_tvt\': .05, # use a small proportion for training\n        \'batch_size\': 32,\n        \'max_epochs\': 11\n    }\n    run(configuration, Data, LitClassifier)\n```\n\n![MNIST_minimal](./mnist_minimal.png)\n\n### Using exposed logger\n\n```python\nfrom plkit import logger\n\n# Your logic\n\nif __name__ == "__main__":\n    # Now you are able to log something outside Modules\n    logger.info(\'Fantastic starts ...\')\n    # ...\n    logger.info(\'Pipeline done.\')\n```\n\n### Best practice (Boilerplate)\n\nYou can use each objects (trainer, module, datamodule) as the way you do with `pytorch-lightning`, but we suggest you use them in the following ways:\n\n#### Running locally\n\n```python\nfrom plkit import Module, DataModule, run\n\nclass MyData(DataModule):\n    ...\n\nclass MyModel(Module):\n    ...\n\nif __name__ == \'__main__\':\n    config = {\n        ...\n    }\n    run(config, MyData, MyModel)\n```\n\n#### Running via SGE\n\n```python\nfrom plkit import Module, DataModule, SGERunner, run\n\n# MyData and MyModel definitions\n\nif __name__ == \'__main__\':\n    config = {\n        ...\n    }\n    sge = SGERunner(...)\n    run(config, MyData, MyModel, runner=sge)\n    # or\n    sge.run(config, MyData, MyModel)\n```\n\n#### With optuna\n\n```python\nfrom plkit import (\n    Module,\n    DataModule,\n    Optuna,\n    OptunaSuggest,\n    LocalRunner,\n    run\n)\n\n# MyData and MyModel definitions\n\nif __name__ == \'__main__\':\n    config = {\n        ...\n        hparam1=OptunaSuggest(<default>, <type>, *args, **kwargs),\n        hparam2=OptunaSuggest(<default>, <type>, *args, **kwargs),\n    }\n    runner = LocalRunner()\n    # requires `val_acc` to be logged in `validation_epoch_end`\n    optuna = Optuna(on=\'val_acc\', n_trials=10, direction=\'maximize\')\n    run(config, MyData, MyModel, runner=runner, optuna=optuna)\n    # or\n    runner.run(config, MyData, MyModel, optuna)\n```\n\n## Resources\n\nSee more examples in `./examples` and [full documentation][2]\n\n\n[1]: https://github.com/PyTorchLightning/pytorch-lightning\n[2]: https://pwwang.github.com/plkit\n[3]: https://img.shields.io/pypi/v/plkit?style=flat-square\n[4]: https://pypi.org/project/plkit/\n[5]: https://img.shields.io/github/v/tag/pwwang/plkit?style=flat-square\n[6]: https://github.com/pwwang/plkit\n[7]: https://img.shields.io/codacy/grade/60b6b06fa06c4539b4a7b48ba30fd2bb?style=flat-square\n[8]: https://app.codacy.com/gh/pwwang/plkit\n[9]: https://img.shields.io/codacy/coverage/60b6b06fa06c4539b4a7b48ba30fd2bb?style=flat-square\n[10]: https://img.shields.io/github/workflow/status/pwwang/plkit/Build%20Docs?label=docs&style=flat-square\n[11]: https://img.shields.io/github/workflow/status/pwwang/plkit/Build%20and%20Deploy?style=flat-square\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pwwang/plkit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
